import std;

// Build a rotation matrix around the X axis, from an angle.
void MatrixBuildRotateX(float4 result[4], float4 angle)
{
  MatrixIdentity(result);

  // Calculate sin and cos for angle.x and put the result in sincos.x and .y
  float4 sincos;
  AngleSinCos(angle, sincos);
  
  result[1].y = sincos.y;
  result[1].z = -sincos.x;

  result[2].y = sincos.x;
  result[2].z = sincos.y;
}

// Build a rotation matrix around the Y axis, from an angle.
void MatrixBuildRotateY(float4 result[4], float4 angle)
{
  MatrixIdentity(result);

  // Calculate sin and cos for angle.x and put the result in sincos.x and .y
  float4 sincos;
  AngleSinCos(angle, sincos);
  
  result[0].x = sincos.y;
  result[0].z = sincos.x;

  result[2].x = -sincos.x;
  result[2].z = sincos.y;
}

// Build a rotation matrix around the Z axis, from an angle.
void MatrixBuildRotateZ(float4 result[4], float4 angle)
{
  MatrixIdentity(result);

  // Calculate sin and cos for angle.x and put the result in sincos.x and .y
  float4 sincos;
  AngleSinCos(angle, sincos);

  result[0].x = sincos.y;
  result[0].y = sincos.x;

  result[1].x = -sincos.x;
  result[1].y = sincos.y;
}

// Build a rotation matrix from euler angles in Y X Z order.
void MatrixBuildRotateXYZ(float4 result[4], float4 euler)
{
  float4 rotateX[4];
  MatrixBuildRotateX(rotateX, euler);

  euler = mr32(euler);
  float4 rotateY[4];
  MatrixBuildRotateY(rotateY, euler);
  
  MatrixMultiply(result, rotateX, rotateY);

  euler = mr32(euler);
  float4 rotateZ[4];
  MatrixBuildRotateZ(rotateZ, euler);
  
  MatrixMultiply(result, result, rotateZ);
}

// Take a vertex in xy:{-1, 1} z:{0, 1} range and transform it into a more GS friendly format
// where xy:{0.0f, 4096.0f} z:{0, max_depth_bits}. It will also homogenize the vertex.
// It assumes that scales = {2048.0f, 2048.0f, max_depth_bits, ?}
void ViewportTransformV(float4 vertex, float4 texCoord, float4 scales)
{
  float q = 1.0f / vertex.w;
  
  // Map the xy range from {-1.0, 1.0} to {-2048.0f, 2048.0f} and then add {2048.0f, 2048.0f}.
  vertex.xyz = scales * float4(vertex.w) + vertex * scales;
  
  // We also need to homogenize the vertex.
  vertex.xyz = vertex * float4(q);
  
  // Translate from the native (floating point) format to GS friendly (fixed point) format.
  vertex = VertexFpToGsXYZ2(vertex);
}

// Also prepares the texCoord for perspective correct texture mapping.
void ViewportTransformVT(float4 vertex, float4 texCoord, float4 scales)
{
  float q = 1.0f / vertex.w;
  
  // Map the xy range from {-1.0, 1.0} to {-2048.0f, 2048.0f} and then add {2048.0f, 2048.0f}.
  vertex.xyz = scales * float4(vertex.w) + vertex * scales;
  
  // We also need to homogenize the vertex and texCoord. (?)
  vertex.xyz = vertex * float4(q);
  texCoord.xyz = texCoord * float4(q);
  
  // Translate from the native (floating point) format to GS friendly (fixed point) format.
  vertex = VertexFpToGsXYZ2(vertex);
}
