#include <stdio.h>
#include <math.h>

#include "defines.h"
#include "vec4.h"
#include "mat44.h"
#include "packet.h"
#include "marker.h"

// Our texture in C header format.
#include "texture.h"

// Floating point constants.
#define kPi               3.1415926535f

// Convert between degrees and radians and vice versa.
#define DegToRad(deg)     ( deg * (kPi / 180.0f) )
#define RadToDeg(rad)     ( rad * (180.0f / kPi) )


// Convert from Rgb24 (8:8:8) into Rgba16 (5:5:5:1) and drop alpha bit.
#define Rgb16(r, g, b)    ( (r >> 3) | ((g >> 3) <<  5) | ((b >> 3) << 10) )

// Number of vertices in the cube.
#define kNumVertices      36

// Mesh geometry data.
Vec4 texCoords[kNumVertices] = 
{	
  Vec4(0, 0, 1, 1), Vec4(1, 0, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 1, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(1, 0, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 1, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(1, 0, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 1, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(1, 0, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 1, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(1, 0, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 1, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(1, 0, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 1, 1, 1), Vec4(0, 1, 1, 1), Vec4(1, 0, 1, 1) 
};

Vec4 normals[kNumVertices] =
{
  Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1),
  Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1), Vec4(0, 0, 1, 1) 
};

Vec4 vertices[kNumVertices] = 
{	
  Vec4(-0.5f,-0.5f, 0.5f, 1), Vec4( 0.5f,-0.5f, 0.5f, 1), Vec4(-0.5f, 0.5f, 0.5f, 1), Vec4( 0.5f, 0.5f, 0.5f, 1), Vec4(-0.5f, 0.5f, 0.5f, 1), Vec4( 0.5f,-0.5f, 0.5f, 1),
  Vec4(-0.5f,-0.5f,-0.5f, 1), Vec4(-0.5f, 0.5f,-0.5f, 1), Vec4( 0.5f,-0.5f,-0.5f, 1), Vec4( 0.5f, 0.5f,-0.5f, 1), Vec4( 0.5f,-0.5f,-0.5f, 1), Vec4(-0.5f, 0.5f,-0.5f, 1),
  Vec4(-0.5f,-0.5f,-0.5f, 1), Vec4(-0.5f,-0.5f, 0.5f, 1), Vec4(-0.5f, 0.5f,-0.5f, 1), Vec4(-0.5f, 0.5f, 0.5f, 1), Vec4(-0.5f, 0.5f,-0.5f, 1), Vec4(-0.5f,-0.5f, 0.5f, 1),
  Vec4( 0.5f,-0.5f,-0.5f, 1), Vec4( 0.5f, 0.5f,-0.5f, 1), Vec4( 0.5f,-0.5f, 0.5f, 1), Vec4( 0.5f, 0.5f, 0.5f, 1), Vec4( 0.5f,-0.5f, 0.5f, 1), Vec4( 0.5f, 0.5f,-0.5f, 1),
  Vec4(-0.5f, 0.5f, 0.5f, 1), Vec4( 0.5f, 0.5f, 0.5f, 1), Vec4(-0.5f, 0.5f,-0.5f, 1), Vec4( 0.5f, 0.5f,-0.5f, 1), Vec4(-0.5f, 0.5f,-0.5f, 1), Vec4( 0.5f, 0.5f, 0.5f, 1),
  Vec4(-0.5f,-0.5f, 0.5f, 1), Vec4(-0.5f,-0.5f,-0.5f, 1), Vec4( 0.5f,-0.5f, 0.5f, 1), Vec4( 0.5f,-0.5f,-0.5f, 1), Vec4( 0.5f,-0.5f, 0.5f, 1), Vec4(-0.5f,-0.5f,-0.5f, 1) 
};

u_long128   packet_data[1024];
gifpacket_t gifpacket;

// Create a perspective transform matrix.
Mat44 BuildPerspective(float vFov, float aspect, float hither = 1.0f, float yon = 1000.0f)
{
  float y = hither * tanf(vFov / 2.0f),
        x = y * aspect,

        n = hither,   // near
        f = yon,      // far
        l = -x,       // left
        r =  x,       // right
        t =  y,       // top
        b = -y;       // bottom

  Mat44 perspective;
  perspective.SetRow(0, Vec4((2 * n) / (r - l),    0,                   0,                        0));
  perspective.SetRow(1, Vec4(0,                   -(2 * n) / (t - b),   0,                        0));
  perspective.SetRow(2, Vec4(((r + l) / (r - l)),  ((t + b) / (t - b)), ((f + n) / (f - n)),     -1));
  perspective.SetRow(3, Vec4(0,                    0,                   ((2 * f * n) / (f - n)),  0));

  return perspective;
}

// Convert from standard number to the corresponding log2 value.
int Log2(int x)
{
  switch (x)
  {
  case    1:  return  0;
  case    2:  return  1;
  case    4:  return  2;
  case    8:  return  3;
  case   16:  return  4;
  case   32:  return  5;
  case   64:  return  6;
  case  128:  return  7;
  case  256:  return  8;
  case  512:  return  9;
  case 1024:  return 10;
  }

  // Assert?
  return -1;
}

// Upload a 16bit texture to GS memory.
void LoadTexture16(int width, int height, u_int *tex, int dstMem)
{
  // Prepare the texture upload by setting the transfer registers.
  gifpacket_addgsdata(&gifpacket, GIF_SET_TAG(NLOOP(4), EOP(0), PRE(0), 0, GIF_PACKED, NREG(1)));
  gifpacket_addgsdata(&gifpacket, AD_REG);

  // Set Destination buffer in BITBLTBUF
  qword_t reg;
  reg.ul64[0] = GS_SET_BITBLTBUF(0, 0, 0,   // Src fields ignored when transfering from host to local.
                                 dstMem,    // Dst memory address (word address / 64).
                                 1,         // Dst buffer width divided by 64.
                                 PSMCT16);  // Pixel format.
  reg.ul64[1] = BITBLTBUF_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  // Set Destination position in TRXPOS.
  reg.ul64[0] = GS_SET_TRXPOS(0, 0,         // x, y for source area. Ignored for host -> GS.
                              0, 0,         // x, y for destination area.
                              0);           // Pixel transmission order. Ignored if not GS -> GS.
  reg.ul64[1] = TRXPOS_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  // Set Destination transmission area in TRXREG
  reg.ul64[0] = GS_SET_TRXREG(width, height);
  reg.ul64[1] = TRXREG_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  // Kick off Transfer with TRXDIR
  reg.ul64[0] = GS_SET_TRXDIR(TRXDIR_TO_GS);
  reg.ul64[1] = TRXDIR_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  // Upload the texture data.
  // Each 16bit pixel is two bytes and each qword is 16 bytes.
  int size = width * height * 2 / 16;
  gifpacket_addgsdata(&gifpacket, GIF_SET_TAG(NLOOP(size), EOP(0), PRE(0), 0, GIF_IMAGE, NREG(0)));
  gifpacket_addgsdata(&gifpacket, 0); // Not used in GIF_IMAGE mode.

  for (int i = 0; i < width * height * 3; i += 8 * 3)
  {
    // Convert from 24-bit to 16-bit color and pack it up into a 128 bit register.
    // This means that we can fit 8 pixels/colors into one register.
    qword_t color;
    color.ui32[0] = Rgb16(tex[i +  0], tex[i +  1], tex[i +  2]) | (Rgb16(tex[i +  3], tex[i +  4], tex[i +  5]) << 16);
    color.ui32[1] = Rgb16(tex[i +  6], tex[i +  7], tex[i +  8]) | (Rgb16(tex[i +  9], tex[i + 10], tex[i + 11]) << 16);
    color.ui32[2] = Rgb16(tex[i + 12], tex[i + 13], tex[i + 14]) | (Rgb16(tex[i + 15], tex[i + 16], tex[i + 17]) << 16);
    color.ui32[3] = Rgb16(tex[i + 18], tex[i + 19], tex[i + 20]) | (Rgb16(tex[i + 21], tex[i + 22], tex[i + 23]) << 16);

    gifpacket_addgspacked(&gifpacket, color.ul128);
  }

  // Bind this texture.
  gifpacket_addgsdata(&gifpacket, GIF_SET_TAG(NLOOP(4), EOP(1), PRE(0), 0, GIF_PACKED, NREG(1)));
  gifpacket_addgsdata(&gifpacket, AD_REG);

  // Access the TEXFLUSH register to tell the GS to finish uploading the texture so we can use it.
  reg.ul64[0] = GS_SET_TEXFLUSH();
  reg.ul64[1] = TEXFLUSH_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  // Set texture properties.
  reg.ul64[0] = GS_SET_TEX0(dstMem,         // Memory address to find the texture.
                            1,              // Buffer width times 64 bytes.
                            PSMCT16,        // Format.
                            Log2(width),    // Width in power of two.
                            Log2(height),   // Height in power of two.
                            0,              // Texture color component (RGB).
                            0,              // Texture function (Modulate).
                            0, 0, 0, 0, 0); // CLUT / Palette.
  reg.ul64[1] = TEX0_1_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  reg.ul64[0] = GS_SET_TEX1(1,              // LOD calculation method (fixed value, LOD = K).
                            0,              // Maximum Mip level.
                            1,              // MAG filter set to linear.
                            1,              // MIN filter set to linear.
                            0,              // Base address spec for mip maps.
                            0,              // LOD parameter L.
                            0);             // LOD parameter K.
  reg.ul64[1] = TEX1_1_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);

  reg.ul64[0] = GS_SET_TEXA(0,              // Alpha value to use when alpha bit in Rgba16 is 0.
                            0,              // If RGB black should be treated as transparent.
                            0);             // Alpha value to use when alpha bit in Rgba16 is 1.
  reg.ul64[1] = TEXA_REG;
  gifpacket_addgspacked(&gifpacket, reg.ul128);
}

int main(int argc, char **argv)
{
  // Reset the data buffer and initialize the packet functions.
  int i;
  for (i = 0; i < 1024; i++)
    packet_data[i] = 0;
  gifpacket_init(&gifpacket, packet_data);

  // Upload the view matrix.
  Marker(&gifpacket, "kViewMatrix");
  Mat44 worldToCamera;
  worldToCamera.LoadIdentity();
  worldToCamera.SetRow(3, Vec4(0, 0, -100.0f, 1));
  gifpacket_addgsmat44(&gifpacket, &worldToCamera);

  // Upload the perspective matrix.
  Marker(&gifpacket, "kPerspectiveMatrix");
  Mat44 cameraToScreen = BuildPerspective(DegToRad(45.0f), 350.0f / 640.0f);
  gifpacket_addgsmat44(&gifpacket, &cameraToScreen);

  // Scales.
  Marker(&gifpacket, "kScales");
  qword_t scales;
  scales.f32[0] = 2048.0f;                      // Horizontal center of the screen.
  scales.f32[1] = 2048.0f;                      // Vertical center of the screen.
  scales.f32[2] = (float)((2 << 24) - 1) / 2;   // Depth buffer resolution.
  scales.f32[3] = 1.0f;
  gifpacket_addgspacked(&gifpacket, scales.ul128);

  // Upload the modelspace data. (Pre-transformed geometry).
  MarkerConst("kMeshNumVertices", kNumVertices);
  Marker(&gifpacket, "kMeshVertices");
  for (i = 0; i < kNumVertices; i++)
  {
    gifpacket_addgsvec4(&gifpacket, &(texCoords[i]));
    gifpacket_addgsvec4(&gifpacket, &(normals[i]));
    gifpacket_addgsvec4(&gifpacket, &(vertices[i]));
  }

  // Set the GIF tag for rendering the mesh.
  Marker(&gifpacket, "kMeshRenderTag");
  int prim = GS_SET_PRIM(GS_PRIM_TRI, GOURAUD(1), TEXTURED(1), FOG(0), ABE(0), ALPHA(0), AA(0), CTXT(0), FIX(0));
  gifpacket_addgsdata(&gifpacket, GIF_SET_TAG(NLOOP(kNumVertices), EOP(1), PRE(1), prim, GIF_PACKED, NREG(3)));
  gifpacket_addgsdata(&gifpacket, 0x512); // { texCoord, color, vertex }

  // Upload the screenspace data. (Post-transformed geometry).
  qword_t color;
  for (i = 0; i < kNumVertices; i++)
  {
    // Leave a gap for transformed STQ co-ords.
    gifpacket_addgspacked(&gifpacket, 0);

    // Add color. Texture modulated with vertex colors requires color components to be 7 bits instead of 8.
    color.ui32[0] = 128; color.ui32[1] = 128; color.ui32[2] = 128; color.ui32[3] = 128;
    gifpacket_addgspacked(&gifpacket, color.ul128);
    
    // Leave a gap for transformed vertex co-ords.
    gifpacket_addgspacked(&gifpacket, 0);
  }

  // First free GS memory area when running the harness.
  int dstMem = 0x118 << 5;
  Marker(&gifpacket, "kTextureUploadTag");
  LoadTexture16(kTexWidth, kTexHeight, tex, dstMem);

  // Finish the gif packet stream.
  gifpacket_terminate(&gifpacket);

  // Save the data stream to disc.
  if (argc != 2)
  {
    fprintf(stderr, "Usage: gen_data output_file\n");
    return -1;
  }

  FILE *output = fopen(argv[1], "wb");
  if (output == NULL)
  {
    fprintf(stderr, "Could not open file \"%s\"\n", argv[1]);
    return -2;
  }

  fwrite(packet_data, sizeof(packet_data) / sizeof(qword_t), sizeof(qword_t), output);
  int bytes = ftell(output);
  fclose(output);
  printf("Written %d bytes to \"%s\"\n", bytes, argv[1]);
  return 0;
}
