
#include <unistd.h>
#include <stdio.h>
#include <linux/ps2/ee.h>
#include "packet.h"

void gifpacket_init(gifpacket_t *gpkt, u_long128 *buffer)
{
	gpkt->base=align128((u_int*)buffer);
	gpkt->current=(u_int*)gpkt->base;
	gpkt->dmatag=NULL;
}

u_long128 *gifpacket_terminate(gifpacket_t *gpkt)
{
	u_long128* const dma_current = align128(gpkt->current);
	if(gpkt->dmatag)
	{
		u_int const count = (dma_current - gpkt->dmatag) - 1;
		((u_int*)gpkt->dmatag)[0] += count;
	}
	gpkt->current = (u_int*)dma_current;
	gpkt->dmatag = NULL;
	return dma_current;
}

int  gifpacket_size(gifpacket_t *gpkt)
{
	return ((u_int)gpkt->current - (u_int)gpkt->base) / sizeof(u_long128);
}

void gifpacket_addgsdata(gifpacket_t *gpkt, u_long64 data)
{
	*((u_long64*)gpkt->current) = data;
	gpkt->current+=sizeof(u_long64)/sizeof(u_int);
}

void gifpacket_addgspacked(gifpacket_t* gpkt, u_long128 data)
{
	*((u_long128*)gpkt->current) = data;
	gpkt->current += sizeof(u_long128) / sizeof(u_int);
}

void gifpacket_addgsvec4(gifpacket_t* gpkt, Vec4* vec)
{
	qword_t data;
	data.f32[0] = vec->x;
	data.f32[1] = vec->y;
	data.f32[2] = vec->z;
	data.f32[3] = vec->w;

	gifpacket_addgspacked(gpkt, data.ul128);
}

void gifpacket_addgsmat44(gifpacket_t* gpkt, Mat44* mat)
{
	Vec4 temp = mat->GetRow(0);
	gifpacket_addgsvec4(gpkt, &temp);
	temp = mat->GetRow(1);
	gifpacket_addgsvec4(gpkt, &temp);
	temp = mat->GetRow(2);
	gifpacket_addgsvec4(gpkt, &temp);
	temp = mat->GetRow(3);
	gifpacket_addgsvec4(gpkt, &temp);
}
