import std;

void TransformVertex(float4 vertexAddr[], float4 saveAddr[], float4 objectToScreen[4],
                     float4 oldVector, float4 oldVertex, int test, float4 scale, float4 lights[4],
                     const int winding)
{
  // Get untransformed vertex data from memory.
  float4 st     = vertexAddr[0];
  float4 normal = vertexAddr[1];
  float4 vertex = vertexAddr[2];

  // Increments vertexAddr, this is a bit dodgy methinks, should be done in loop outside.
  vertexAddr += 3;

  // Transform to screen position.
  vertex = MatrixMultiplyVertex(objectToScreen, vertex);

  // Setup the clipping.
  // Using the intrinsics generates 1 more instruction but use no more steps overall, due to pipelining (in final asm).
  clipW(vertex);
  int clipflag = fcand(0x3FFFF);
  clipflag += 0x7FFF;
  
  // Homogenize the vertex coordinate.
  float q     = 1.0f / vertex.w;
  vertex.xyz  = vertex * float4(q);
  st.xyz      = st * float4(q);

  // Offset the vertex for the PS2 screen space.
  vertex.xyz = scale * vertex + scale;

  int adc = int(st.w);

  float4 vector;
  vector.xyz = vertex - oldVertex;

  // Is the result of these ops a flag setting? That is tested with fmand?
  float4 crossproduct;
  if (winding == 1)
    crossproduct = CrossProduct(oldVector, vector);
  else
    crossproduct = CrossProduct(vector, oldVector);

  int result  = fmand(test);
  result      += 0xFFE0;
  adc         = adc | result;

  oldVector.xyz = vector;
  oldVertex.xyz = vertex;

  // Convert the position to fixed point integer.
  vertex = ftoi4(vertex);

  // Set the clip flag in the vertex.
  adc = adc | clipflag;
  vertex.w = float(adc);

  // Calculate the final vertex colour from the lights.
  float4 ints = MatrixMultiplyVector(lights, normal);
  // Note: the asm version only used vf00, i.e. the w would be 1 does this matter?
  ints = max(ints, float4(0.0f));
  ints *= lights[3];

  float4 vertexColour = ftoi0(esum(ints));

  // Store the final transformed vertex back into memory for rendering.
  saveAddr[0].xyz = st;
  saveAddr[1].xyz = vertexColour.xyz;
  saveAddr[2]     = vertex;
  saveAddr        += 3;
}

void main(float4 vuMem[])
{
  // Read the TOP register (for double buffering).
  float4 dataAddr[];
  xtop(dataAddr);

  // Get global transformation, scale and light information from memory.
  float4 objectToScreen[4] = dataAddr;
  float4 lights[4]         = dataAddr + 5;
  float4 scale;
  scale.xyz = dataAddr[4].xyz;

  // Reset clipping by setting the clip flag to zero.
  fcset(0);

  // Initialize to default values.
  float4  oldVector = float4(0.0f); // ToDo: const regs? vf00;
  float4  oldVertex = oldVector;    // vf00;
  int     test      = 0x0020;

  // Loop forever.
  for (;;)
  {
    // Tell the double buffering that this is the continue point when the program is restarted.
    asm { --cont }

    // Read the TOP register (for double buffering).
    xtop(dataAddr);
    dataAddr[256] = dataAddr[0];

    // Transform the vertices.
    float4 primAddr[]      = dataAddr + 256;
    float4 vertexAddr[]    = dataAddr + 1;
    float4 maxvertexAddr[] = dataAddr + 253;
    float4 saveAddr[]      = dataAddr + 257;
    do
    {
      TransformVertex(vertexAddr, saveAddr, objectToScreen, oldVector, oldVertex, test, scale, lights, 0);

      // Second vertex (alternate winding).
      TransformVertex(vertexAddr, saveAddr, objectToScreen, oldVector, oldVertex, test, scale, lights, 1);
    }
    while (vertexAddr != maxvertexAddr);

    // Draw the geometry to screen.
    xgkick(primAddr);
  }
}
